// dissociated-press implementation for the command line
// donated into the public domain by Chris Lightfoot, 1997

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

// useful defines
#define PROGNAME "dissociate"
#define ERROR(text)		fprintf(stderr, PROGNAME ": " text "\n");

// random defines
#define MAX_SHORT		0xffff	// assume 16 bits
#define MAX_WORDLENGTH		256
#define INITIAL_BUFFERLEN	16384
#define DEFAULT_RUNLENGTH	10
#define DEFAULT_MATCHES		2
#define DEFAULT_PRINTWORDS	-1
#define DEFAULT_LINELENGTH	70
#define MIN_LINELENGTH		20

// long random number
int lrand(void);

int lrand()
{
	// Hmm....
	return rand();
/*
	int l;

	l = rand();
	l |= rand() << 15;

	return l;
*/
}

// hash code for a word
unsigned short hash(char *s);

unsigned short hash(char *s)
{
	unsigned short h = 0;
	char *p;

	for (p = s; *p; p++)
	{
		// this is pretty crap (and only good for 16 letters, but hey).
		h ^= (unsigned long)*p;
		h = h << 1;
	}

	return h;
}

// the big table of words in the text
struct tagdictrecord
{
	unsigned long h;	// hash code for this word
	char *word;
	struct tagdictrecord *next;
};

typedef struct tagdictrecord dictrecord;

dictrecord **dictionary;

// the internal representation of the text itself
dictrecord **text;

// read a word from the standard input, returning zero if we reach end-of-file
int getword(char *s);

int getword(char *s)
{
	// this is pretty shit but should work
	return fscanf(stdin, "%s ", s) == 1;
}

// we read from stdin, write to stdout
int main(int argc, char **argv)
{
	// flag for whether we print out information (number of words found)
	int debugmode = 0;
	int numwords = 0, numdictentries = 0;
	// things to do with text, dictionary.
	char s[MAX_WORDLENGTH];
	unsigned short h;
	dictrecord *pd, **ppt, **pptnew;
	// for managing the text buffer.
	long wordsallocated;
	// features of the output
	int nummatches, runlength, printwords;
	// used in the main `generating' loop
	int i, j, k, l, direction, locflag;
	// used to print out text more `prettily'
	int x, linelength;

	// here we check through the parameters
	// --help		print a help string
	// --matches n		number of matching words required
	// --runlength n	how often to jump
	// --printwords n	total number of words to print out.
	// --seed n		seed random number generator
	// --linelength n	length of lines to print out
	// --debug		print debugging information
	// skip past program name
	k = time(NULL);
	nummatches = DEFAULT_MATCHES;
	runlength = DEFAULT_RUNLENGTH;
	printwords = DEFAULT_PRINTWORDS;
	linelength = DEFAULT_LINELENGTH;

	for (i = 1; i < argc; i++)
	{
		if (!strcmp(argv[i], "--help"))
		{
			fprintf(stderr,
				   "An implementation of `dissociated press' text mangling\n"
				   "\n"
				   "Command line options:\n"
				   "    --help            print this help message\n"
				   "    --debug           print debugging information on standard output\n"
				   "    --seed n          set random number seed to n\n"
				   "    --matches n       require n matching words after each jump\n"
				   "    --runlength n     print n words between jumps\n"
				   "    --printwords n    print n words of output in total\n"
				   "    --linelength n    wraps words to n characters long\n"
				   "\n"
				   "donated into the public domain by Chris Lightfoot <chris@ex-parrot.com>\n");

			return 5;
		}

		if (!strcmp(argv[i], "--debug"))
			debugmode = 1;

		if (!strcmp(argv[i], "--seed") && i < (argc - 1))
			k = atoi(argv[i + 1]);

		if (!strcmp(argv[i], "--runlength") && i < (argc - 1))
		{
			runlength = atoi(argv[i + 1]);
			if (runlength == 0) runlength = DEFAULT_RUNLENGTH;
		}

		if (!strcmp(argv[i], "--matches") && i < (argc - 1))
		{
			nummatches = atoi(argv[i + 1]);
			// we impose an arbitrary upper limit on nummatches
			if (nummatches < 1 || nummatches > runlength)
				nummatches = DEFAULT_MATCHES;
		}
		if (!strcmp(argv[i], "--printwords") && i < (argc - 1))
		{
			printwords = atoi(argv[i + 1]);
			if (printwords == 0) printwords = DEFAULT_PRINTWORDS;
		}

		if (!strcmp(argv[i], "--linelength") && i < (argc - 1))
		{
			linelength = atoi(argv[i + 1]);
			if (linelength < MIN_LINELENGTH && linelength != -1) linelength = DEFAULT_LINELENGTH;
		}
	}

	srand(k);

	if (debugmode)
	{
		fprintf(stderr, "used random seed %d\n", k);
		fprintf(stderr, "using run length %d, number of matches %d, length of output text %d\n",
						runlength, nummatches, printwords);
		if (linelength != -1)
			fprintf(stderr, "wrapping words to lines length %d characters\n", linelength);
		else fprintf(stderr, "not wrapping words");
	}

	// allocate space for the dictionary
	dictionary = (dictrecord**) malloc(MAX_SHORT * sizeof(dictrecord*));
	if (!dictionary)
	{
		ERROR("failed to allocate memory for dictionary");
		return 1;
	}

	for (i = 0; i < MAX_SHORT; i++)
		dictionary[i] = NULL;

	// also allocate some space for the text
	text = (dictrecord**) malloc((wordsallocated = INITIAL_BUFFERLEN) * sizeof(dictrecord*));
	if (!text)
	{
		ERROR("failed to allocate memory for text");
		return 2;
	}

	ppt = text;
	*ppt = NULL;

	// the main input loop: we read in words, adding them to the list of words
	// in the text (and to the dictionary, if necessary);
	while (getword(s)) // fetch a word
	{
		// now we look up the dictionary record for s
		h = hash(s);

		if (pd = dictionary[h])
			for (; pd && strcmp(pd->word, s); pd = pd->next);

		if (pd == NULL)
		{
			// now add record to the dictionary.
			pd = (dictrecord*)malloc(sizeof(dictrecord));
			if (pd == NULL)
			{
				ERROR("failed to allocate memory for a dictionary record");
				return 6;
			}
			pd->word = (char*)malloc(strlen(s) + 1);
			if (pd->word == NULL)
			{
				ERROR("failed to allocate memory for a dictionary record");
				return 6;
			}
			strcpy(pd->word, s);
			pd->h = h;
			pd->next = dictionary[h];
			dictionary[h] = pd;

			numdictentries++;
		}

		// now we have a dictionary record and we add a reference to it to
		// the end of the text
		*ppt = pd;
		ppt++; numwords++;

		// check whether we are about to overrun the buffer
		if ((ppt - text) == (wordsallocated - 1))
		{
			i = ppt - text;
			
			// increase available storage 50%
			wordsallocated += wordsallocated / 2;
			
			text = (dictrecord**)realloc((void*)text, wordsallocated * sizeof(dictrecord*));

			if (text == NULL)
			{
				ERROR("failed to allocate additional storage for large input buffer");
				return 3;
			}

			ppt = text + i;
		}
	}

	*ppt = NULL;

	if (numwords <= nummatches)
	{
		ERROR("number of required matches was greater than text length");
		return 4;
	}

	if (debugmode)
	{
		fprintf(stderr, "text was %d words long\n", numwords);
		fprintf(stderr, "found %d distinct words in input text\n", numdictentries);
	}

	// now we choose a random place to start, and then start printing stuff randomly
	ppt = text + (lrand() % numwords);
	i = 0;	// total number of words printed

	x = 0;

	do
	{
		// print out that set of words
		for (j = 0; j < runlength; i++, j++, ppt++)
		{
			// if we've reached the end of the text, we begin again somewhere else
			if (*ppt == NULL)
			{
				ppt = text + (lrand() % numwords);
				printf("\n\n");
			}

			// word wrapping
			if (linelength != -1)
				if ((x + strlen((*ppt)->word) + 1) >= linelength)
				{
					x = 0;
					printf("\n");
				}
			
			printf("%s ", (*ppt)->word);
			x += strlen((*ppt)->word) + 1;
		}

		// now pick a new location to start from
		l = 0;
		
		do
		{
			pptnew = text + (lrand() % numwords);
			direction = (rand() & 1) ? 1 : -1;

// considered harmful?
test:
			pptnew += direction;
			
			for (k = 1, locflag = 1; k <= nummatches; k++)
				if (*(pptnew - k) != *(ppt - k)) locflag = 0;

			if (!locflag && pptnew > (text + k) && *pptnew) goto test;

			l++;
		}
		while (!locflag);

		ppt = pptnew;
	}
	while (printwords == -1 || i < printwords);

	// finish off the output with a CRLF
	printf("\n");

	return 0;
}
